// saratogaConsole.cpp - provide input/output to a saratoga daemon thread
//
// see LICENSE.txt for licensing information
//
// (c) 2008 Wesley M. Eddy
// 

// normal C/C++ stuff
#include <stdio.h>

#ifdef _MSC_VER
// Windows OS-specific stuff
#include <process.h>
#else
#include <pthread.h>
#endif

// local project files
#include "commands.h"

const wchar_t *version = L"0.1.0";

bool killAllThreads = false;

std::wstring lineEntered; // this needs to be protected from access by multiple threads
#ifdef _MSC_VER
HANDLE gLineEnteredEvent;
HANDLE gLineEnteredMutex;
HANDLE gCmdHandled;
#else
pthread_cond_t gLineEnteredEvent;
pthread_mutex_t gLineEnteredEventMutex;
pthread_mutex_t gLineEnteredMutex;
pthread_cond_t gCmdHandled;
pthread_mutex_t gCmdHandledMutex;
#endif


#ifdef _MSC_VER
DWORD WINAPI getLinesFromStdin(__in LPVOID) {
#else
void * getLinesFromStdin(void *) {
#endif
	wprintf(L"Wesley Eddy's Saratoga Implementation (version %ls)\n%ls",
		version, promptString1);

	const size_t MAX_CMD_LEN = 1024;
	wchar_t buffer[MAX_CMD_LEN];
	while (!killAllThreads) {
		if (fgetws(buffer, MAX_CMD_LEN, stdin) != NULL) {
			// grab mutex, copy buffer, and signal to other
			// thread that does processing
            #ifdef _MSC_VER
			WaitForSingleObject(gLineEnteredMutex, INFINITE);
            #else
            pthread_mutex_lock(&gLineEnteredMutex);
            #endif
			lineEntered = std::wstring(buffer);
            // strip newline
            lineEntered.erase(lineEntered.length()-1);
            // signal processing thread that input is available
            #ifdef _MSC_VER
			ReleaseMutex(gLineEnteredMutex);
			SetEvent(gLineEnteredEvent);
			WaitForSingleObject(gCmdHandled, INFINITE);
			ResetEvent(gCmdHandled);
            #else
            pthread_mutex_unlock(&gLineEnteredMutex);
            pthread_mutex_lock(&gLineEnteredEventMutex);
            pthread_cond_signal(&gLineEnteredEvent);
            pthread_mutex_unlock(&gLineEnteredEventMutex);
            pthread_mutex_lock(&gCmdHandledMutex);
            pthread_cond_wait(&gCmdHandled, &gCmdHandledMutex);
            pthread_mutex_unlock(&gCmdHandledMutex);
            #endif
		}
	}
	return 0;
}

#ifdef _MSC_VER
DWORD WINAPI runSaratoga(__in LPVOID) {
#else
void * runSaratoga(void *) {
#endif
	saratoga::startup();
    saratoga::log_printf("started saratoga thread\n");
	while (!killAllThreads) {
        int activity = 0;
        activity += saratoga::processIncomingPackets() ? 1 : 0;
        activity += saratoga::processOutgoingPackets() ? 1 : 0;
        activity += saratoga::processEvents() ? 1 : 0;
        if (activity == 0) ::Sleep(1); // don't kill the CPU
	}
    saratoga::log_printf("cleaning up ...\n");
	saratoga::cleanup();
	return 0;
}

int main(int argc, char* argv[])
{
	// create event for signaling command received from stdin
    #ifdef _MSC_VER
	gLineEnteredEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
	if (gLineEnteredEvent == NULL) {
		printf("CreateEvent failed (%d)\n", GetLastError());
		return 1;
	}
    #else
    int sysRet;
    sysRet = pthread_mutex_init(&gLineEnteredEventMutex, NULL);
    if (sysRet != 0) {
        printf("pthread_mutex_init failed (%d)\n", sysRet);
        return 1;
    }
    sysRet = pthread_cond_init(&gLineEnteredEvent, NULL);
    if (sysRet != 0) {
        printf("pthread_cond_init failed (%d)\n", sysRet);
        return 1;
    }
    #endif
	// command processed event
    #ifdef _MSC_VER
	gCmdHandled = CreateEvent(NULL, TRUE, false, NULL);
	if (gCmdHandled == NULL) {
		printf("CreateEvent failed (%d)\n", GetLastError());
		return 2;
	}
    #else
    sysRet = pthread_mutex_init(&gCmdHandledMutex, NULL);
    if (sysRet != 0) {
        printf("pthread_mutex_init failed (%d)\n", sysRet);
        return 2;
    }
    sysRet = pthread_cond_init(&gCmdHandled, NULL);
    if (sysRet != 0) {
        printf("pthread_cond_init failed (%d)\n", sysRet);
        return 2;
    }
    #endif

	// create mutex for controlling access to string holding
	// entered command line
    #ifdef _MSC_VER
	gLineEnteredMutex = CreateMutex(NULL, false, NULL);
	if (gLineEnteredMutex == NULL) {
		printf("CreateMutex failed (%d)\n", GetLastError());
		return 3;
	}
    #else
    if ((sysRet = pthread_mutex_init(&gLineEnteredMutex, NULL)) != 0) {
        printf("pthread_mutex_init failed (%d)\n", sysRet);
        return 3;
    }
    #endif

	// fire off thread for collecting input
    #ifdef _MSC_VER
	HANDLE inputThread;
	inputThread = CreateThread(NULL, 0, &getLinesFromStdin, NULL, 0, NULL);
	if (inputThread == NULL) {
		printf("CreateThread failed (%d)\n", GetLastError());
		return 4;
	}
    #else
    pthread_t inputThread;
    sysRet = pthread_create(&inputThread, NULL, &getLinesFromStdin, NULL);
    if (sysRet != 0) {
        printf("pthread_create failed (%d)\n", sysRet);
        return 4;
    }
    #endif

	// fire off thread for actually doing Saratoga protocol work
    #ifdef _MSC_VER
	HANDLE saratogaThread;
	saratogaThread = CreateThread(NULL, 0, &runSaratoga, NULL, 0, NULL);
	if (saratogaThread == NULL) {
		printf("CreateThread failed (%d)\n", GetLastError());
		return 5;
	}
    #else
    pthread_t saratogaThread;
    sysRet = pthread_create(&saratogaThread, NULL, &runSaratoga, NULL);
    if (sysRet != 0) {
        printf("pthread_create failed (%d)\n", sysRet);
        return 5;
    }
    #endif

	// process any input packets or commands
	while (!killAllThreads) {
        #ifdef _MSC_VER
		if (WaitForSingleObject(gLineEnteredEvent, 1) == WAIT_OBJECT_0) {
        #else
        pthread_mutex_lock(&gLineEnteredEventMutex);
        pthread_cond_wait(&gLineEnteredEvent, &gLineEnteredEventMutex);
        pthread_mutex_unlock(&gLineEnteredEventMutex);
        #endif
			// grab mutex to read string
            #ifdef _MSC_VER
			WaitForSingleObject(gLineEnteredMutex, INFINITE);
            #else
            pthread_mutex_lock(&gLineEnteredMutex);
            #endif
			handleCommand(lineEntered);
			lineEntered.clear();
            #ifdef _MSC_VER
			ReleaseMutex(gLineEnteredMutex);
			ResetEvent(gLineEnteredEvent);
			SetEvent(gCmdHandled);
            #else
            pthread_mutex_unlock(&gLineEnteredMutex);
            pthread_mutex_lock(&gCmdHandledMutex);
            pthread_cond_signal(&gCmdHandled);
            pthread_mutex_unlock(&gCmdHandledMutex);
            #endif
        #ifdef _MSC_VER
		}
        #endif
	}

    #ifdef _MSC_VER
    const DWORD threadDieTimeout = 10*1000;
    WaitForSingleObject(inputThread, threadDieTimeout);
    WaitForSingleObject(saratogaThread, threadDieTimeout);
    #else
    pthread_join(inputThread, NULL);
    pthread_join(saratogaThread, NULL);
    #endif

	// cleanup
    #ifdef _MSC_VER
	CloseHandle(gLineEnteredEvent);
	CloseHandle(gLineEnteredMutex);
	CloseHandle(gCmdHandled);
    #else
    pthread_cond_destroy(&gLineEnteredEvent);
    pthread_mutex_destroy(&gLineEnteredEventMutex);
    pthread_mutex_destroy(&gLineEnteredMutex);
    pthread_cond_destroy(&gCmdHandled);
    pthread_mutex_destroy(&gCmdHandledMutex);
    #endif
	return 0;
}

