// saratoga.cpp - core logic for Saratoga reference library
//
// see LICENSE.txt for licensing information
//
// (c) 2008 Wesley M. Eddy
// 

#include "saratoga.h"

namespace saratoga {

// for rate control
unsigned int packetPeriod = 10; // ms between DATA packets

std::map<std::wstring, peerStruct> peers;
std::map<std::string, transactionStruct> transactions;

LPCWSTR unconfiguredEID = L"*EID Unconfigured*";
std::wstring localEID = unconfiguredEID;
void setLocalEID(std::wstring &eid) {
	localEID = eid;
}

void listPeers(std::list<std::wstring> &l) {
	l.clear();
    std::map<std::wstring, peerStruct>::iterator i;
	for (i = peers.begin(); i != peers.end(); i++) {
		const size_t maxAddr = 1024;
		char addr[maxAddr];
		inet_ntop(
			i->second.saddr.ss_family,
			((i->second.saddr.ss_family == AF_INET) ?
				&(((struct sockaddr_in *)&(i->second.saddr))->sin_addr) :
				((i->second.saddr.ss_family == AF_INET6) ?
					&(((struct sockaddr_in6 *)&(i->second.saddr))->sin6_addr) :
					(void *)&(i->second.saddr))),
			addr, maxAddr);
		sprintf(addr+strlen(addr), " : %u",
			htons(((struct sockaddr_in *)&(i->second.saddr))->sin_port));
		l.push_back(i->first + std::wstring(L" @ ") + wstringFromUtf8(addr, strlen(addr)));
	}
}

void addPeer(std::wstring &EID, struct sockaddr_storage *sa) {
    const size_t convStrLen = 1024;
    char convStr[convStrLen];
    const char *r;
    r = inet_ntop(sa->ss_family,
                  sa->ss_family == AF_INET ? (void *)&((struct sockaddr_in *)sa)->sin_addr
                                           : (void*)&((struct sockaddr_in6 *)sa)->sin6_addr,
                  convStr, convStrLen);
    if (r == NULL) strcpy(convStr, "(invalid)");
    std::string eidStr = stringFromWstring(EID.c_str(), EID.length());
    log_printf("adding peer EID: %s, IP: %s, port: %u\n", eidStr.c_str(), convStr,
               ntohs(sa->ss_family == AF_INET ? ((struct sockaddr_in *)sa)->sin_port
                                               : ((struct sockaddr_in6 *)sa)->sin6_port));
	peerStruct p;
	p.EID = EID;
    memcpy(&p.saddr, sa, sizeof(struct sockaddr_storage));
	peers[EID] = p;
}

void addPeer(std::wstring &EID, std::string &ipAddr, uint16_t port) {
    struct sockaddr_storage addr;
	if (!(inet_pton(AF_INET, ipAddr.c_str(), &addr) > 0)) {
		if (!(inet_pton(AF_INET6, ipAddr.c_str(), &addr) > 0)) {
			printf("bad IP address format! (%s)\n", ipAddr.c_str());
			return;
		} else { // XXX IPv6 stuff
			((struct sockaddr_in6 *)&addr)->sin6_port = htons(port);
		}
	} else { // IPv4-specific stuff
		((struct sockaddr_in *)&addr)->sin_port = htons(port);
	}
    addPeer(EID, &addr);
}

bool sendBeaconPackets = true;
DWORD beaconPeriod = 5000, // time between BEACONs sent (in milliseconds)
      lastBeaconSentAt = 0; // time the last BEACON was sent

void printGetdirResult(transactionStruct *t) {
	wprintf(L"Remote Directory Listing for %s\n\n", t->dirent.filePath.c_str());
	fflush(t->fp);
	fseek(t->fp, SEEK_SET, 0);
	const short MAX_DIRENT = 4096;
	char buf[MAX_DIRENT];
	direntStruct d;
	while (fread(buf, 1, 13, t->fp) == 13 && fgets(buf+13, MAX_DIRENT-13, t->fp) != NULL) {
		decodeDirent(&d, buf, MAX_DIRENT);
		wprintf(L"%s\t%llu bytes\t%u\t%u\t%u\n", d.filePath.c_str(), d.size, d.ctime, d.mtime, d.properties);
	}
}

void transactionCompleted(transactionStruct *t) {
	if (t->fp != NULL) {
		if (t->direction == RECEIVER && t->type == SARATOGA_GETDIR) { // if it was a getdir, print the result
			printGetdirResult(t);
		}

		if (fclose(t->fp) == 0) t->fp = NULL;
		if (t->direction == RECEIVER && (t->type == SARATOGA_GET || t->type == SARATOGA_GETDIR)) {
			sendHolestofill(t);
		}
	}
}

void transactionAborted(transactionStruct *t) {
	transactionCompleted(t);
}

bool processEvents() {
    bool eventsProcessed = false;
	DWORD now = GetTickCount();

	std::list<std::string> completedTransactionIDs;
	completedTransactionIDs.clear();

	// for all transactions, see if we can send anything, if they've timed out, etc.
	std::map<std::string, transactionStruct>::iterator it = transactions.begin();
	while (it != transactions.end()) {
		transactionStruct *t = &(it->second);

		// check if time to send HOLESTOFILL
		// XXX

		// look for METADATA to send
		if ((it->second.direction == SENDER) &&
		    ((!(it->second.metadataSent == true)) ||
			 ((!(it->second.metadataReceived == true)) &&
			  (now - it->second.metadataSentTime > RETRANSMIT_TIMER)))) {
			sendMetadata(&(it->second));
            eventsProcessed = true;
		}

		// check for completed transactions
		if ((it->second.metadataReceived == true) &&
			((it->second.cumack >= it->second.dirent.size))) // XXX should look at size of holeList too
		{
			transactionCompleted(&(it->second));
			// transactions are erased from the transactions map later in
			// order to avoid issues with the current iteration
			completedTransactionIDs.push_back(it->first);
			it++;
            eventsProcessed = true;
			continue;
		}

		// is it timed out?
		if (now - t->lastReceivedTime > TIMEOUT_MS) {
			printf("closing transaction 0x%08x due to inactivity\n", it->second.transactionID);
			// XXX
			transactionAborted(&(it->second));
			completedTransactionIDs.push_back(it->first);
			it++;
            eventsProcessed = true;
			continue;
		}

		// see if REQUEST needs to be retransmitted
		if (t->lastReceivedTime == 0 && (now - t->lastSentTime > RETRANSMIT_TIMER)) {
			sendRequest(&(it->second));
			it++;
            eventsProcessed = true;
			continue;
		}

		// is there data to be sent?
		if (t->direction == SENDER && (now >= t->nextSendTime)) {
			sendData(&(it->second));
            eventsProcessed = true;
		}

		it++;
	}

	std::list<std::string>::iterator ctit = completedTransactionIDs.begin();
	while (ctit != completedTransactionIDs.end()) {
		transactions.erase(*ctit);
		ctit++;
	}

	// check if it's time for BEACONs to be sent
    if (sendBeaconPackets && now - lastBeaconSentAt > beaconPeriod) {
        sendBeacon();
        eventsProcessed = true;
    }
    return eventsProcessed;
}

FILE *logFP = NULL;
void log_printf(const char* formatstr, ...) {
	const size_t maxStr = 4096;
	if (logFP == NULL) {
        const char *logfile = "packetLog.txt";
		logFP = fopen(logfile, "w");
        if (logFP == NULL) printf("unable to open %s\n", logfile);
	}
    char timeStr[maxStr];
#ifdef _MSC_VER
    __time64_t now;
    _time64(&now);
    sprintf(timeStr, "%u - ", now);
#else
    time_t now = time(NULL);
    struct tm *ts = localtime(&now);
    strftime(timeStr, sizeof(timeStr), "%a %H:%M:%S %Z ", ts);
#endif

	va_list	argptr;
	char str[maxStr];
	va_start(argptr, formatstr);
	vsprintf(str, formatstr, argptr);
	va_end(argptr);
	if (logFP == NULL || fprintf(logFP, timeStr) < 0 || fprintf(logFP, str) < 0)
    {
        printf("error printing to log file! (%s)\n", str);
	}
}



#ifdef _MSC_VER
uint32_t filetimeToDtnTime(FILETIME *f) {
	uint64_t t = (((uint64_t)(f->dwHighDateTime))<<32) + f->dwLowDateTime; // move 2 32-bit values into 1 64-bit
	t /= 10000000000; // convert nanoseconds to seconds
	const uint64_t conv = 12591482400; // (2000 - 1601) * 365.25 * 24 * 60 * 60
	t -= conv; // convert Jan 1 1601 to Y2K epoch
	return (uint32_t)(t & 0xFFFFFFFF);
}
std::string ListDir(LPCWSTR dirname) {
	BOOL fFinished;
	HANDLE hList;
	WCHAR szDir[MAX_PATH+1];
	WIN32_FIND_DATA FileData;
	std::string ret;
	saratoga::direntStruct d;

	// Get the proper directory path
	swprintf(szDir, L"%s\\*", dirname);
	struct _stat sbuf;

	// Get the first file
	hList = FindFirstFile(szDir, &FileData);
	if (hList == INVALID_HANDLE_VALUE)	{ 
		log_printf("No files found\n\n");
	}
	else {
		// Traverse through the directory structure
		fFinished = FALSE;
		while (!fFinished) {
			d.filePath = FileData.cFileName;
			d.ctime = filetimeToDtnTime(&FileData.ftCreationTime);
			d.mtime = filetimeToDtnTime(&FileData.ftLastWriteTime);
			if (_wstat(d.filePath.c_str(), &sbuf) == -1) return ret; // error; stat needed to detect "special" files

			// Check the object is a directory or not
			if (FileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
				if ((wcscmp(FileData.cFileName, L".") != 0) &&
					(wcscmp(FileData.cFileName, L"..") != 0)) {
					// filter out . and .. entries
				} else {
					d.properties = saratoga::SARATOGA_DIRECTORY_PROPERTY;
					d.size = 0;
					ret += saratoga::encodeDirent(&d);
				}
			}
			//else if (S_ISFIFO(&sbuf)) { // stream
			else if ((sbuf.st_mode & S_IFREG) != S_IFREG) { // assume stream XXX
				d.properties = saratoga::SARATOGA_STREAM_PROPERTY;
				d.size = 0;
				ret += saratoga::encodeDirent(&d);
			}
			else { // normal file
				d.properties = 0;
				// windows bizarreness
				d.size = (((uint64_t)FileData.nFileSizeHigh) << 32) + FileData.nFileSizeLow;
				if (FileData.nFileSizeLow < 0)
					d.size += 0x100000000;

				ret += saratoga::encodeDirent(&d);
			}

			if (!FindNextFile(hList, &FileData)) {
				if (GetLastError() == ERROR_NO_MORE_FILES) {
					fFinished = TRUE;
				}
			}
		}
	}

	FindClose(hList);
	return ret;
}

#else

#include <dirent.h>

std::string ListDir(const wchar_t *wdir) {
    DIR *dp;
    struct dirent *dir_entry;
    struct stat stat_info;
    std::string ret;
    saratoga::direntStruct d;
    std::string dirStr = stringFromWstring(wdir, wcslen(wdir));
    const char *dir = dirStr.c_str();

    if((dp = opendir(dir)) == NULL) {
        fprintf(stderr,"cannot open directory: %s\n", dir);
        return "";
    }
    chdir(dir);
    while((dir_entry = readdir(dp)) != NULL) {
        d.properties = 0;
        lstat(dir_entry->d_name,&stat_info);
        if(S_ISDIR(stat_info.st_mode)) {
            /* Directory, but ignore . and .. */
            if(strcmp(".",dir_entry->d_name) == 0 ||
                strcmp("..",dir_entry->d_name) == 0)
                continue;
            d.properties = saratoga::SARATOGA_DIRECTORY_PROPERTY;
        }
        d.filePath = wstringFromUtf8(dir_entry->d_name, strlen(dir_entry->d_name));
        d.size = stat_info.st_size;
        d.ctime = stat_info.st_ctime;
	d.mtime = stat_info.st_mtime;
        ret += saratoga::encodeDirent(&d);
    }
    chdir("..");
    closedir(dp);
    return ret;
}

#endif

// adjust seconds from UNIX epoch of Jan 1, 1970 to DTN epoch of Jan 1, 2000
uint32_t unixEpochToDtnEpoch(uint32_t t) {
  return t - 946728000; /* 30 * 365.25 * 24 * 60 * 60 */
}

void buildDirent(transactionStruct *t, requestStruct *r) {
#ifdef _MSC_VER
	struct _stat sbuf;
	if (_fstat(_fileno(t->fp), &sbuf) == -1) return;
#else
	struct stat sbuf;
	if (fstat(fileno(t->fp), &sbuf) == -1) return;
#endif

	t->dirent.filePath = r->remoteFilePath;
	t->dirent.ctime = unixEpochToDtnEpoch(sbuf.st_ctime & 0xFFFFFFFF);
	t->dirent.mtime = unixEpochToDtnEpoch(sbuf.st_mtime & 0xFFFFFFFF);
	t->dirent.properties = 0; // XXX
	t->dirent.size = sbuf.st_size;
}

bool transactionExists(const std::string &k) {
	std::map<std::string, transactionStruct>::iterator it;
	it = transactions.find(k);
	return (it != transactions.end());
}

void initTransaction(transactionStruct *t) {
 	t->lastReceivedTime = GetTickCount();
	t->holeList.clear();
	t->cumack = 0;
	t->nextSendTime = t->lastReceivedTime + packetPeriod;
}

void respondToRequest(requestStruct *r) {
	transactionStruct t;
	t.type = r->type;
	t.transactionID = r->transactionID;
	memcpy(&(t.peer), &(r->peer), sizeof(peerStruct));
	std::string key = transactionKey(&t);

	t.direction = SENDER; // only the sender ever gets REQUESTs (not used in _put_)
	if (transactionExists(key)) {
		// XXX ignore?
		log_printf("ignoring request; transaction already exists\n");
		return;
	}
	if (r->type == SARATOGA_DELETE || r->type == SARATOGA_DELETEDIR) {
		// XXX not yet implemented
		log_printf("_delete_ transactions are not yet implemented\n");
		return;
	}
	else if (r->type == SARATOGA_GETDIR) {
		// try to read the directory entry
		std::string encodedDirents = ListDir(r->remoteFilePath.c_str());
		// write it to file
#ifdef _MSC_VER
		r->localFileName = _wtempnam(L"", L".");
#else
        char tempfile[1024];
        memset(tempfile, 0, 1024);
        mkstemp(tempfile); // XXX check return code
        r->localFileName = wstringFromUtf8(tempfile, strlen(tempfile));
#endif
        std::string strFileName = stringFromWstring(r->localFileName.c_str(), r->localFileName.length());
        if ((t.fp = fopen(strFileName.c_str(), "w+")) == NULL) {
			log_printf("unable to open temporary file for getdir result\n");
			sendHolestofill(&t, SARATOGA_ERROR_UNSPECIFIED);
			return;
		}
		fwrite(encodedDirents.data(), 1, encodedDirents.size(), t.fp);
		fflush(t.fp);
		// fill in METADATA
		buildDirent(&t, r);
		log_printf("built dirent for getdir\n");
	}
	// only type of valid request left is a _GET_
	else if (r->type != SARATOGA_DELETE && r->type != SARATOGA_DELETEDIR) {
		// verify that reading in this directory is allowed
		// XXX
		// open file
        std::string filename = stringFromWstring(r->remoteFilePath.c_str(), r->remoteFilePath.length());
		t.fp = fopen(filename.c_str(), "r+b");
		if (t.fp == NULL) {
			log_printf("unable to open path specified in request\n");
			// notify other side
			sendHolestofill(&t, SARATOGA_ERROR_FILE_NOT_FOUND);
			return;
		}
		// fill in metadata
		buildDirent(&t, r);
	}
	initTransaction(&t);
	// add onto transactions map
    std::string peerEIDstr = stringFromWstring(t.peer.EID.c_str(), t.peer.EID.length());
	log_printf("inserting %s -> %s : %u\n", key.c_str(), peerEIDstr.c_str(), t.transactionID);
    transactions[key] = t;
}

void startGetTransaction(std::wstring *p, std::wstring *remoteFileName,
                         std::wstring *localFileName) {
	transactionStruct t;
	initTransaction(&t);
	
    std::string strFileName = stringFromWstring(localFileName->c_str(), localFileName->length());
    t.fp = fopen(strFileName.c_str(), "w+");
    if (t.fp == NULL) {
        log_printf("ERROR: unable to open local file: %s\n", strFileName.c_str());
        return;
    }
	t.direction = saratoga::RECEIVER;
	t.dirent.filePath = *remoteFileName;
	t.type = SARATOGA_GET;
	t.transactionID = 0; // XXX
    std::map<std::wstring, peerStruct>::iterator pit = peers.find(*p);
	if (pit == peers.end()) {
        wprintf(L"ERROR: peer EID %s not found!\n", p->c_str());
        return;
    }
	memcpy(&(t.peer), &(pit->second), sizeof(peerStruct));
	std::string key = transactionKey(&t);

	// add onto transactions map
	std::pair<std::string, transactionStruct>(key, t);
    std::string peerName = stringFromWstring(t.peer.EID.c_str(), t.peer.EID.length());
	log_printf("inserting %s -> %s : %u\n", key.c_str(), peerName.c_str(), t.transactionID);
	transactions.insert(transactions.begin(), std::make_pair(key, t));

	// send request packet
	sendRequest(&t);
}

/*
std::list<hole>::iterator findMatchedHole(uint32_t dataStart, uint32_t dataEnd,
                                          std::list<hole> &holeList) {
    std::list<struct hole>::iterator ret;
    for (ret = holeList.begin(); ret != holeList.end(); ret++) {
        // check if first byte, at least, within hole
        if (dataStart >= ret->first && dataStart <= ret->second) break;
        // check if last byte, at least, within hole
        if (dataEnd >= ret->first && dataEnd <= ret->second) break;;
        // check if hole lies within packet
        if (dataStart <= ret->first && dataEnd >= ret->second) break;
    }
    return ret;
}
*/

std::wstring wstringFromUtf8(const char *utf8str, size_t utf8strLen) {
    const size_t maxOutput = 1024;
    wchar_t output[maxOutput];
    char *coutput = (char *)output;
    memset(output, 0, sizeof(wchar_t) * maxOutput);
    size_t outputLen = maxOutput;
#ifdef _MSC_VER
    MultiByteToWideChar(CP_UTF8, 0, utf8str, utf8strLen, output, outputLen);
#else
    iconv_t A2W = iconv_open("WCHAR_T", "UTF-8");
    char *cutf8str = (char *)malloc(maxOutput);
    if (cutf8str == NULL) return L"";
    memcpy(cutf8str, utf8str, utf8strLen);
    iconv(A2W, &cutf8str, &utf8strLen, &coutput, &outputLen);
    iconv_close(A2W);
    free(cutf8str);
#endif
    return output;
}

std::string stringFromWstring(const wchar_t *wchars, size_t wcharsLen) {
    const size_t maxOutput = 1024;
    char output[maxOutput] = { 0 };
    size_t outputLen = maxOutput;
#ifdef _MSC_VER
    WideCharToMultiByte(CP_UTF8, 0, wchars, wcharsLen, output, outputLen, NULL, NULL);
#else
    char *cwchars = (char *)wchars;
    char *coutput = output;
    iconv_t W2A = iconv_open("UTF-8", "WCHAR_T");
    wcharsLen *= sizeof(wchar_t); // needs to be in bytes
    iconv(W2A, &cwchars, &wcharsLen, &coutput, &outputLen);
    iconv_close(W2A);
#endif
    return output;
}

};

// end saratoga namespace

#ifndef _MSC_VER
uint32_t GetTickCount() {
    static long hz = sysconf(_SC_CLK_TCK);
    tms tm;
    return times(&tm) * (1000 / hz);
}
#endif
