// saratoga.h - main header for Saratoga reference library
//
// see LICENSE.txt for licensing information
//
// Thanks to Muhammad.Muhammad@dlr.de for some adjustments to aid in
//   compilation on Linux
//
// (c) 2008 Wesley M. Eddy
// 

#ifndef _SARATOGA_H
#define _SARATOGA_H 1

// cross-platform C include files
#include <wchar.h>
#include <time.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>

// platform-specific C includes and definitions
#ifdef _MSC_VER
#include <winsock2.h>
#include <ws2tcpip.h>
#pragma comment(lib,"ws2_32.lib") 
#include <process.h>
typedef unsigned __int8 uint8_t;
typedef unsigned __int16 uint16_t;
typedef unsigned __int32 uint32_t;
typedef unsigned __int64 uint64_t;

typedef int ssize_t;

// for UTF-8 conversion
#include <winnls.h>

#else // linux version
#include <stdint.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <errno.h>

typedef uint32_t DWORD;
typedef wchar_t * LPWSTR;
typedef const wchar_t * LPCWSTR;
typedef char * LPSTR;
typedef const char * LPCSTR;
typedef int WSADATA;
typedef int SOCKET;
const int SOCKET_ERROR = -1;
const int INVALID_SOCKET = 1;

#define closesocket(x) close(x)
#define WSAGetLastError() ((long)errno)
#define WSACleanup(x) {}
#define WSAStartup(x,y) 0

#include <strings.h>
#define Sleep(x) usleep(1000*x)
#define ZeroMemory(x,y) bzero(x,y)
#define _tempnam tempnam
#include <sys/times.h>
uint32_t GetTickCount(void);

// for UTF-8 conversion
#include "iconv.h"
#endif

// cross-platform C++ include files
#include <string>
#include <list>
#include <map>
#include <sys/stat.h>
#include <sys/types.h>

namespace saratoga {

const uint32_t TIMEOUT_MS = 5000; // how long to wait on a silent transaction before timing out
const uint32_t RETRANSMIT_TIMER = 5000; // for retransmitting REQUEST and METADATA if not acked

extern LPCWSTR unconfiguredEID;
extern std::wstring localEID;
void setLocalEID(std::wstring &eid);

extern uint16_t localPort;

void listPeers(std::list<std::wstring> &l);
void addPeer(std::wstring &EID, struct sockaddr_storage *sa);
void addPeer(std::wstring &EID, std::string &ipAddr, uint16_t port);

bool processEvents();

bool processOutgoingPackets(); // return true if packets are actually sent

// enable "server" mode by responding positively to REQUESTs
extern bool serveRequests;

// internally-used stuff
typedef struct peerStruct {
	std::wstring EID;
	struct sockaddr_storage saddr;
} peerStruct;
extern std::map<std::wstring, peerStruct> peers;

extern FILE *logFP;
void log_printf(const char* formatstr, ...);

typedef enum {
	SARATOGA_GET = 0,
	SARATOGA_GETDIR = 1,
	SARATOGA_DELETE = 2,
	SARATOGA_DELETEDIR = 3,
	SARATOGA_PUT = 4, // not a real type in packets
} transactionType;

typedef struct direntStruct {
	uint64_t size;
	uint32_t mtime, ctime;
	uint8_t properties;
	std::wstring filePath;
} direntStruct;
enum fileProperties { // for the properties field in direntStruct
	SARATOGA_DIRECTORY_PROPERTY = 0x01,
	SARATOGA_STREAM_PROPERTY = 0x02
};

typedef std::pair<uint32_t, uint32_t> hole;

enum {
	SENDER, RECEIVER
};

typedef struct transactionStruct {
	transactionType type;
	uint32_t transactionID;
	direntStruct dirent;
	FILE *fp;
	uint32_t lastReceivedTime, lastSentTime;
	bool metadataSent, metadataReceived;
	uint32_t metadataSentTime;
	unsigned int direction; // SENDER or RECEIVER
	peerStruct peer;

	uint32_t cumack; // highest offset contiguously acked from 0
    uint32_t lastDataSent;
	std::list<hole> holeList;

	uint32_t nextSendTime; // for rate control
	bool operator <(const transactionStruct &other) const;
} transactionStruct;

std::string transactionKey(transactionStruct *t);

extern std::map<std::string, transactionStruct> transactions;

// socket stuff in saratogaSocket.cpp
int startup();
void cleanup();
bool processIncomingPackets(); // return true if packets read

// packet sending and receiving stuff from saratogaPackets.cpp
extern bool sendBeaconPackets;
bool sendBeacon();
void sendRequest(transactionStruct *t);
void sendMetadata(transactionStruct *t);
void sendHolestofill(transactionStruct *t, uint8_t errorCode = 0);
void sendData(transactionStruct *t);

void handlePacket(char *packetBuf, size_t packetLen,
	              struct sockaddr_storage *fromAddr);
int decodeDirent(direntStruct *d, char *e, uint16_t eLen);
std::string encodeDirent(direntStruct *d);

typedef struct requestStruct {
	transactionType type;
	uint32_t transactionID;
	std::wstring remoteFilePath, localFileName;
	peerStruct peer;
} requestStruct;

void respondToRequest(requestStruct *r);

enum errors {
	SARATOGA_SUCCESS = 0x00,
	SARATOGA_ERROR_UNSPECIFIED = 0x01,
	SARATOGA_ERROR_FILE_NOT_FOUND = 0x04
};

bool transactionExists(const std::string &k);

void startGetTransaction(std::wstring *p, std::wstring *fname,
                         std::wstring *localFileName);

std::list<hole>::iterator findMatchedHole(uint32_t dataStart, uint32_t dataEnd,
                                          std::list<hole> &holeList);

std::wstring wstringFromUtf8(const char *utf8str, size_t utf8strLen);
std::string stringFromWstring(const wchar_t *wchars, size_t wcharsLen);
};

#endif
