// saratogaPacket.cpp - sending & handling routines for Saratoga packets
//
// see LICENSE.txt for licensing information
//
// (c) 2008 Wesley M. Eddy
// 

#include "saratoga.h"

namespace saratoga {

// from saratogaSocket.cpp
extern SOCKET ipv4UnicastSocket, ipv6UnicastSocket;
extern struct sockaddr_in ipv4MulticastAddress;
extern struct sockaddr_in6 ipv6MulticastAddress;

const uint16_t MAX_PACKET_SIZE = 4096; // buffer size used for single packet receiving

const uint8_t SARATOGA_V1 = 0x40;
enum packetTypes {
	SARATOGA_BEACON_PACKET = 0,
	SARATOGA_REQUEST_PACKET = 1,
	SARATOGA_METADATA_PACKET = 2,
	SARATOGA_DATA_PACKET = 3,
	SARATOGA_HOLESTOFILL_PACKET = 4
};

enum flags {
	FLAG_METADATA_RECEIVED = 0x04, // on HOLESTOFILL packet
	FLAG_REQUEST_HOLESTOFILL = 0x01, // on DATA packet
	FLAG_REQUEST_DIRECTORY = 0x01 // on REQUEST packet for _getdir_ or _delete_ directory
};

// numeric field encoding and decoding routines
static inline uint16_t read16(char *p) {
//	return (((uint16_t)p[0])<<8) | ((uint8_t)p[1]);
    return ntohs(*(uint16_t *)p);
}
static inline uint32_t read32(char *p) {
//	return ((uint32_t)read16(p)<<16) | read16(p+2);
    return ntohl(*(uint32_t *)p);
}
static inline void encode16(char *p, uint16_t x) {
//	p[0] = (x>>8); p[1] = x & 0xFF;
    *(uint16_t *)p = htons(x);
}
static inline void encode32(char *p, uint32_t x) {
//	encode16(p, (x>>16)); encode16(p+2, (x & 0xFFFF));
    *(uint32_t *)p = htonl(x);
}

bool processOutgoingPackets() {
    int packetsSent = 0;
	// XXX nothing to be done here anymore; packets are sent in processEvents
    return packetsSent != 0;
}

peerStruct* findPeerByAddress(struct sockaddr_storage *rcvAddr) {
    std::map<std::wstring, peerStruct>::iterator it;
	for (it = peers.begin(); it != peers.end(); it++)
	{
		if (memcmp(rcvAddr, &(it->second.saddr),
			       sizeof(struct sockaddr_storage)) == 0)
		{
            // ZZZ
			return &(it->second);
		}
	}
	log_printf("findPeerByAddress() not found\n");
	return NULL;
}

std::string transactionKey(transactionStruct *t) {
	const size_t tmpLen = 2048 + sizeof(struct sockaddr_storage);
	char tmp[tmpLen] = {0};
	const size_t maxAddr = 1024;
	char addr[maxAddr];
	inet_ntop(
		t->peer.saddr.ss_family,
		((t->peer.saddr.ss_family == AF_INET) ?
			&(((struct sockaddr_in *)&(t->peer.saddr))->sin_addr) :
			((t->peer.saddr.ss_family == AF_INET6) ?
				&(((struct sockaddr_in6 *)&(t->peer.saddr))->sin6_addr) :
				(void *)&(t->peer.saddr))),
		addr, maxAddr);
	std::string ret = addr;
    sprintf(tmp, "-%u", t->transactionID);
    ret.append(tmp, tmpLen);
	return ret;
}

extern DWORD lastBeaconSentAt;
bool sendBeacon() {
log_printf("in sendBeacon\n");
	char beaconPacket[MAX_PACKET_SIZE] = {0};
	beaconPacket[0] = SARATOGA_V1 | SARATOGA_BEACON_PACKET;
	beaconPacket[1] = beaconPacket[2] = beaconPacket[3] = 0;
	int packetLen = 4;
	if (wcscmp(localEID.c_str(), unconfiguredEID) != 0) {
log_printf("adding eid\n");
        char *beaconEid = beaconPacket+4;
        std::string utf8eid = stringFromWstring(localEID.c_str(), localEID.length());
        memcpy(beaconEid, utf8eid.data(), utf8eid.size());
        size_t eidLen = utf8eid.size();
		beaconPacket[eidLen+4] = '\0';
		packetLen += eidLen+1;
	}
log_printf("sending packet\n");
    // the unicast socket is used for sending through because it's bound to the port selected
	int ret1 = sendto(ipv4UnicastSocket, beaconPacket, packetLen,
                      0, (struct sockaddr *)&ipv4MulticastAddress,
                      sizeof(ipv4MulticastAddress));
	if (ret1 == SOCKET_ERROR) {
		log_printf("sendto() for IPv4 multicast beacon failed: %ld\n", WSAGetLastError());
	}
    int ret2 = sendto(ipv6UnicastSocket, beaconPacket, packetLen,
                      0, (struct sockaddr *)&ipv6MulticastAddress,
                      sizeof(ipv6MulticastAddress));
    if (ret2 == SOCKET_ERROR) {
        log_printf("sendto() for IPv6 multicast beacon failed: %ld\n", WSAGetLastError());
    }
	lastBeaconSentAt = GetTickCount();
	return (ret1 == packetLen) && (ret2 == packetLen);
}

void handleBeaconPacket(char *p, size_t pLen, struct sockaddr_storage *sa) {
	log_printf("received BEACON packet of length %d bytes\n", pLen);
	if (pLen < 4) {
		log_printf("\tpacket too short to be legal!\n");
		return;
	}
	log_printf("\tflags: 0x%x 0x%x 0x%x\n", p[1], p[2], p[3]);
	if (pLen > 4) {
		if (p[pLen-1] == '\0') {
			log_printf("\tendpoint ID: %s\n", p+4);
		} else {
			log_printf("\tnon-null terminated string for endpoint ID\n");
		}
	} else {
		log_printf("no endpoint ID\n");
	}
	// lookup EID in peers list, and add if not found
    // XXX
	std::wstring EID = (pLen > 4) ? wstringFromUtf8(p+4, pLen-4) : L"";
    std::map<std::wstring, peerStruct>::iterator pit;
	pit = peers.find(EID);
	if (pit == peers.end() && EID.length() > 0) {
        addPeer(EID, sa);
	} else if (EID.length() != 0) {
		// check that sockaddr matches in peers list
        if (memcmp(sa, &(peers[EID].saddr), sizeof(struct sockaddr_storage)) != 0) {
            wprintf(L"EID %s has multiple addresses; using oldest\n", EID.c_str());
        }
	}
}

void sendRequest(transactionStruct *t) {
	char requestPacket[MAX_PACKET_SIZE];
	requestPacket[0] = SARATOGA_V1 | SARATOGA_REQUEST_PACKET;
	requestPacket[1] = requestPacket[2] = requestPacket[3] = 0; // XXX no flags?
	if (t->type == SARATOGA_GETDIR) requestPacket[1] |= FLAG_REQUEST_DIRECTORY;
	encode32(requestPacket+4, t->transactionID);
    std::string utf8path = stringFromWstring(t->dirent.filePath.c_str(), t->dirent.filePath.size());
	memcpy(requestPacket+8, utf8path.c_str(), utf8path.length());
	requestPacket[8+t->dirent.filePath.length()] = '\0';
	int requestLen = 8 + t->dirent.filePath.length()+1; // XXX no authentication field

	int ret = sendto(ipv4UnicastSocket, requestPacket, requestLen,
						0, (struct sockaddr *)&(t->peer.saddr),
						sizeof(struct sockaddr_in));
	if (ret == SOCKET_ERROR) {
		log_printf("sendto() for IPv4 unicast Saratoga REQUEST packet failed: %ld\n", WSAGetLastError());
	}
	t->lastSentTime = GetTickCount();
}

bool serveRequests = false;

void handleRequestPacket(char *p, size_t pLen, struct sockaddr_storage *sa) {
    printf("%s : %u\n", inet_ntoa(((struct sockaddr_in *)sa)->sin_addr), ntohs(((struct sockaddr_in *)sa)->sin_port));
	// log packet
	log_printf("received REQUEST packet of length %d bytes\n", pLen);
	if (pLen < 4) {
		log_printf("\tpacket too short to be legal!\n");
		return;
	}
	log_printf("\tflags: 0x%x 0x%x 0x%x\n", p[1], p[2], p[3]);
	if (pLen < 8) {
		log_printf("\tpacket too short to be legal!\n");
		return;
	}
	uint32_t transactionID = read32(p+4);
	log_printf("\tId: 0x%x\n", transactionID);
	log_printf("\tfile path: %s\n", p+8);

	if (!serveRequests) {
		return;
	}

	requestStruct r;
	r.remoteFilePath = wstringFromUtf8(p+8, strlen(p+8));
	peerStruct *peer = findPeerByAddress(sa);
	if (peer == NULL) {
		// XXX
        log_printf("peer not found\n");
        return;
	}
	memcpy(&(r.peer), peer, sizeof(peerStruct));
	r.type = transactionType(p[1] & 0x03);
	r.transactionID = transactionID;
	respondToRequest(&r);
}

const uint16_t defaultSegSize = 1200; // default segment size, in bytes

uint32_t getNextOffsetToSend(transactionStruct *t) {
    const uint32_t dataPacketHeaderSize = 12; // XXX wrong if timestamps are used
    uint32_t ret = t->lastDataSent + (defaultSegSize-dataPacketHeaderSize);
    if (ret >= t->dirent.size) ret = 0;  // wrap
    // XXX currently, ignores t->holes!
    t->lastDataSent = ret;
    return ret;
}

void sendData(transactionStruct *t) {
	// XXX should go into transaction state ?
	uint16_t segSize = defaultSegSize;
	char dataPacket[MAX_PACKET_SIZE];
	dataPacket[0] = SARATOGA_V1 | SARATOGA_DATA_PACKET;
	dataPacket[1] = dataPacket[2] = dataPacket[3] = 0; // XXX no flags?
	encode32(dataPacket+4, t->transactionID);
	// XXX no timestamp

	uint32_t offset = getNextOffsetToSend(t);
	encode32(dataPacket+8, offset);
    if (t->fp == NULL) {
        log_printf("in sendData, file pointer was NULL\n");
        return;
    }
	fseek(t->fp, offset, SEEK_SET); // XXX check for error
	size_t bytesRead = fread(dataPacket+12, 1, segSize - 12, t->fp); // XXX check return
	int dataLen = bytesRead+12;

	// set request HOLESTOFILL if this is the final DATA packet
	if (offset+bytesRead >= t->dirent.size) {
		dataPacket[1] |= FLAG_REQUEST_HOLESTOFILL;
	}

	int ret = sendto(ipv4UnicastSocket, dataPacket, dataLen,
						0, (struct sockaddr *)&(t->peer.saddr),
						sizeof(struct sockaddr_in));
	if (ret == SOCKET_ERROR) {
		log_printf("sendto() for IPv4 unicast Saratoga DATA packet failed: %ld\n", WSAGetLastError());
	}
}

void handleDataPacket(char *p, size_t pLen, struct sockaddr_storage *sa) {
	log_printf("received DATA packet of length %d bytes\n", pLen);
	if (pLen < 8) {
		log_printf("\tpacket too short to be legal!\n");
		return;
	}
	log_printf("\tflags: 0x%x 0x%x 0x%x\n", p[1], p[2], p[3]);
	uint32_t transactionID = read32(p+4);
	log_printf("\ttransaction ID: 0x%08x\n", transactionID);
	// XXX no timestamp
	uint32_t offset = read32(p+8),
	         lastByte = offset + pLen - 12;
	log_printf("\toffset: 0x%08x , last byte: 0x%08x\n", offset, lastByte);

	// lookup transaction and see if I'm the sender
	transactionStruct tp;
	peerStruct *thePeer = findPeerByAddress(sa);
	if (thePeer == NULL) {
		return; // XXX
	}
	memcpy(&tp.peer, thePeer, sizeof(peerStruct));
	tp.transactionID = transactionID;
	std::string key = transactionKey(&tp);

	if (!(transactionExists(key))) {
		log_printf("\ttransaction unknown\n");
		return;
	} else if (transactions[key].direction == SENDER) {
		log_printf("sender should not get a DATA packet!\n");
		return;
	}

    transactionStruct &t = transactions[key];
// XXX
    if (offset != t.cumack) return; // forces in-order processing of DATA
//    std::list<hole>::iterator affectedHole = findMatchedHole(offset, lastByte, t.holeList);
//    if (affectedHole == t.holeList.end()) return; // data in packet not needed

    bool updateHoles = false; // assume failure of fseek/fwrite until proven otherwise
	int r = fseek(t.fp, offset, SEEK_SET);
    if (r == 0) {
		if (fwrite(p+12, 1, pLen-12, t.fp) == (uint16_t)(pLen - 12)) {
	        updateHoles = true;
		}
    } else {
        log_printf("no fwrite; fseek returned $d\n", r);
    }

/*
    if (updateHoles) {
        // see if packet is at beginning of affected hole
        bool beginsHole = affectedHole->first == offset;
        // see if packet is at end of affected hole
        bool endsHole = affectedHole->second == lastByte;
        // either hole goes away entirely
        // XXX
        // or hole is reduced in size from front
        // XXX
        // or hole is reduced in size from back
        // XXX
        // or more than one hole is touched ...
        // XXX
    }
*/
}

std::string encodeDirent(direntStruct *d) {
	char tmp[MAX_PACKET_SIZE];
	ZeroMemory(tmp, MAX_PACKET_SIZE);
    std::string filePath = stringFromWstring(d->filePath.c_str(), d->filePath.length());
	log_printf("encoding dirent (size: %llu, mtime %u, ctime %u, filePath %s)\n", d->size, d->mtime, d->ctime, filePath.c_str());
	// XXX only ever use 32-bit descriptor ...
	encode32(tmp, d->size);
	encode32(tmp+4, d->mtime);
	encode32(tmp+8, d->ctime);
	tmp[12] = d->properties;
	memcpy(tmp+13, d->filePath.c_str(), d->filePath.length());
	return std::string(tmp, 13+d->filePath.length()+1);
}

void sendMetadata(transactionStruct *t) {
	log_printf("sending a METADATA packet\n");
	char metadataPacket[MAX_PACKET_SIZE];
	metadataPacket[0] = SARATOGA_V1 | SARATOGA_METADATA_PACKET;
	metadataPacket[1] = metadataPacket[2] = metadataPacket[3] = 0; // XXX no flags?
	encode32(metadataPacket+4, t->transactionID);
	std::string direntStr = encodeDirent(&(t->dirent));
	memcpy(metadataPacket+8, direntStr.data(), direntStr.size());
	int metadataLen = 8 + direntStr.size();

	log_printf("sending metadata of length %u\n", metadataLen);

	int ret = sendto(ipv4UnicastSocket, metadataPacket, metadataLen,
						0, (struct sockaddr *)&(t->peer.saddr),
						sizeof(struct sockaddr_in));
	if (ret == SOCKET_ERROR) {
		log_printf("sendto() for IPv4 unicast Saratoga METADATA packet failed: %ld\n", WSAGetLastError());
	}
	t->metadataSentTime = GetTickCount();
	t->metadataSent = true;
}

int decodeDirent(direntStruct *d, char *e, uint16_t eLen) {
	d->size = read32(e);
	d->mtime = read32(e+4);
	d->ctime = read32(e+8);
	d->properties = e[8];
	d->filePath = wstringFromUtf8(e+12, eLen-12);
	return 12+d->filePath.length()+1;
}

void handleMetadataPacket(char *p, size_t pLen, struct sockaddr_storage *sa) {
	log_printf("received METADATA packet of length %d bytes\n", pLen);
	if (pLen < 8) {
		log_printf("\tpacket too short to be legal!\n");
		return;
	}
	log_printf("\tflags: 0x%x 0x%x 0x%x\n", p[1], p[2], p[3]);
	uint32_t transactionID = read32(p+4);
	log_printf("\ttransaction ID: 0x%08x\n", transactionID);

	// XXX find transaction and make sure I'm receiver
	transactionStruct tp;
	peerStruct *peer = findPeerByAddress(sa);
	if (peer == NULL) {
		// XXX
	}
	memcpy(&(tp.peer), peer, sizeof(peerStruct));
	tp.transactionID = transactionID;
	std::string key = transactionKey(&tp);

	if (!(transactionExists(key))) {
		log_printf("\ttransaction unknown\n");
		return;
	} else if (transactions[key].direction == SENDER) {
		log_printf("\tsender should not see metadata!\n");
		return;
	}

	decodeDirent(&(transactions[key].dirent), p+8, pLen-8);
	log_printf("\tfile path: %s\n", transactions[key].dirent.filePath.c_str());
	log_printf("\tfile size: %llu\n", transactions[key].dirent.size);
	log_printf("\tfile mtime: %u\n", transactions[key].dirent.mtime);
	log_printf("\tfile ctime: %u\n", transactions[key].dirent.ctime);
	log_printf("\tfile properties: %u\n", transactions[key].dirent.properties);

	transactions[key].metadataReceived = true;
	
	if (p[1] & FLAG_REQUEST_HOLESTOFILL) {
		sendHolestofill(&transactions[key]);
	}
}

void sendHolestofill(transactionStruct *t, uint8_t errorCode) {
	char htfPacket[MAX_PACKET_SIZE];
	htfPacket[0] = SARATOGA_V1 | SARATOGA_HOLESTOFILL_PACKET;
	htfPacket[1] = htfPacket[2] = htfPacket[3] = 0; // XXX no flags?
	htfPacket[1] |= t->metadataReceived ? FLAG_METADATA_RECEIVED : 0;
	encode32(htfPacket+4, t->transactionID);

	encode32(htfPacket+8, t->cumack);
	encode32(htfPacket+12, 0); // in-response-to
	int htfLen = 16;

	if (errorCode != 0) htfPacket[3] = errorCode;
	else {
		// XXX actually fill in holes list ...
	}

	int ret = sendto(ipv4UnicastSocket, htfPacket, htfLen,
						0, (struct sockaddr *)&(t->peer.saddr),
						sizeof(struct sockaddr_in));
	if (ret == SOCKET_ERROR) {
		log_printf("sendto() for IPv4 unicast Saratoga DATA packet failed: %ld\n", WSAGetLastError());
	}
}

void handleHolestofillPacket(char *p, size_t pLen, struct sockaddr_storage *sa) {
	log_printf("received HOLESTOFILL packet of length %d bytes\n", pLen);
	if (pLen < 16) {
		log_printf("\tpacket too short to be legal!\n");
		return;
	}
	log_printf("\tflags: 0x%x 0x%x\n", p[1], p[2]);
	uint8_t status = p[3];
	uint32_t transactionID = read32(p+4);
	printf("\ttransaction ID: 0x%08x\n", transactionID);
	uint32_t cumack = read32(p+8);
	printf("\tcumack: 0x%08x\n", cumack);
	uint32_t inResponseTo = read32(p+12);
	printf("\tin-response-to descriptor: 0x%08x\n", inResponseTo);

	uint16_t numHoles = (pLen - 12) / 8;
	printf("\t%u holes\n", numHoles);

	// lookup transaction and see if I'm the sender
	transactionStruct tp;
	peerStruct *peer = findPeerByAddress(sa);
	if (peer == NULL) {

	}
	memcpy(&(tp.peer), peer, sizeof(peerStruct));
	tp.transactionID = transactionID;
	std::string key = transactionKey(&tp);

	if (!(transactionExists(key))) {
		printf("\ttransaction unknown\n");
		return;
	} else if (transactions[key].direction != SENDER) {
		if (status != SARATOGA_SUCCESS) {
			// other side encountered an error; should remove transaction
			printf("HOLESTOFILL status code %u\n", status);
			transactions.erase(key);
			return;
		}
		printf("receiver side should not get HOLESTOFILL with 0 status code\n");
		return;
	}

	if (transactions[key].cumack > cumack) return; // possible reordering in network ACK path
	transactions[key].cumack = cumack;

	if ((p[1] | FLAG_METADATA_RECEIVED) == FLAG_METADATA_RECEIVED) {
		transactions[key].metadataReceived = true;
	}
}

void handlePacket(char *packetBuf, size_t packetLen,
                  struct sockaddr_storage *fromAddr) {
	// sanity check on version number field
	if ((packetBuf[0] & 0xC0) != SARATOGA_V1) {
		log_printf("read a non Saratoga v1 packet!\n");
		return;
	}

	switch (packetBuf[0] & 0x3F) {
		case SARATOGA_BEACON_PACKET:
			handleBeaconPacket(packetBuf, packetLen, fromAddr);
			break;
		case SARATOGA_REQUEST_PACKET:
			handleRequestPacket(packetBuf, packetLen, fromAddr);
			break;
		case SARATOGA_DATA_PACKET:
			handleDataPacket(packetBuf, packetLen, fromAddr);
			break;
		case SARATOGA_METADATA_PACKET:
			handleMetadataPacket(packetBuf, packetLen, fromAddr);
			break;
		case SARATOGA_HOLESTOFILL_PACKET:
			handleHolestofillPacket(packetBuf, packetLen, fromAddr);
			break;
		default:
			log_printf("unknown Saratoga packet type (%d)\n", packetBuf[0]&0x3f);
	}
	
}

}
