// saratogaSocket.cpp - code for dealing with socket interface
//
// see LICENSE.txt for licensing information
//
// (c) 2008 Wesley M. Eddy
// 

#include "saratoga.h"

namespace saratoga {

const uint16_t SARATOGA_DEFAULT_PORT = 7542; // allocated from IANA
const char SARATOGA_IPV4_MULTICAST_ADDR[] = "224.0.0.108"; // allocated from IANA
const char SARATOGA_IPV6_MULTICAST_ADDR[] = "FF02:0:0:0:0:0:0:6c"; // allocated from IANA

WSADATA wsaData;
SOCKET ipv4UnicastSocket = SOCKET_ERROR, ipv4MulticastSocket = SOCKET_ERROR,
       ipv6UnicastSocket = SOCKET_ERROR, ipv6MulticastSocket = SOCKET_ERROR;
struct sockaddr_in ipv4MulticastAddress;
struct sockaddr_in6 ipv6MulticastAddress;
uint16_t localPort = SARATOGA_DEFAULT_PORT;

int createUdpSocket(SOCKET *s, int type) {
	*s = socket(type, SOCK_DGRAM, IPPROTO_UDP);
	if (*s == INVALID_SOCKET) {
		log_printf("socket() failed: %ld\n", WSAGetLastError());
        return -1;
	}
    return 0;
}

int setSocketReuseAddr(SOCKET *s) {
    #ifdef _MSC_VER
	    char on = 1;
    #else
	    int on = 1;
    #endif
	int ret = setsockopt(*s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	if (ret == SOCKET_ERROR) { // non-fatal to our program
		log_printf("setsockopt() for SO_REUSEADDR failed: %ld\n", WSAGetLastError());
        return -1;
    }
    return 0;
}

int startup() {
	int ret;

	// winsock initialization
	ret = WSAStartup(MAKEWORD(2,2), &wsaData);
	if (ret != 0) {
		log_printf("WSAStartup() failed with %d\n", ret);
		return 1;
	}

	// create UDP sockets
    if (createUdpSocket(&ipv4UnicastSocket, AF_INET) != 0 ||
        createUdpSocket(&ipv4MulticastSocket, AF_INET) != 0 ||
        createUdpSocket(&ipv6UnicastSocket, AF_INET6) != 0 ||
        createUdpSocket(&ipv6MulticastSocket, AF_INET6) != 0)
    {
		cleanup();
		return 2;
	}

	// set reuse-address/port socket option
    if (setSocketReuseAddr(&ipv4UnicastSocket) != 0 ||
        setSocketReuseAddr(&ipv4MulticastSocket) != 0 ||
        setSocketReuseAddr(&ipv6UnicastSocket) != 0 ||
        setSocketReuseAddr(&ipv6MulticastSocket) != 0)
    {
        // non-fatal 
    }

	// bind sockets to appropriate ports
    struct sockaddr_in ipv4Address;
    struct sockaddr_in6 ipv6Address;
	ZeroMemory(&ipv4Address, sizeof(ipv4Address));
	ipv4Address.sin_family = AF_INET;
	ipv4Address.sin_port = htons(localPort);
	ipv4Address.sin_addr.s_addr = htonl(INADDR_ANY);
	ret = bind(ipv4UnicastSocket, (struct sockaddr *)&ipv4Address,
				sizeof(ipv4Address));
	if (ret == SOCKET_ERROR) {
		log_printf("bind() for IPv4 unicast failed: %ld\n", WSAGetLastError());
		cleanup();
		return 4;
	}
    ZeroMemory(&ipv6Address, sizeof(ipv6Address));
    ipv6Address.sin6_family = AF_INET6;
    ipv6Address.sin6_port = htons(localPort);
    ipv6Address.sin6_addr = in6addr_any;
    ret = bind(ipv6UnicastSocket, (struct sockaddr *)&ipv6Address,
               sizeof(ipv6Address));
    if (ret == SOCKET_ERROR) {
        log_printf("bind() for IPv6 unicast failed: %ld\n", WSAGetLastError());
        cleanup();
        return 5;
    }
	// always use default port for multicast receiving
    memcpy(&ipv4MulticastAddress, &ipv4Address, sizeof(ipv4Address));
	ipv4MulticastAddress.sin_port = htons(SARATOGA_DEFAULT_PORT);
	ret = bind(ipv4MulticastSocket, (struct sockaddr *)&ipv4MulticastAddress,
				sizeof(ipv4MulticastAddress));
	if (ret == SOCKET_ERROR) {
		log_printf("bind() for IPv4 multicast failed: %ld\n", WSAGetLastError());
		cleanup();
		return 6;
	}
    memcpy(&ipv6MulticastAddress, &ipv6Address, sizeof(ipv6Address));
    ipv6MulticastAddress.sin6_port = htons(SARATOGA_DEFAULT_PORT);
    ret = bind(ipv6MulticastSocket, (struct sockaddr *)&ipv6MulticastAddress,
               sizeof(ipv6MulticastAddress));
    if (ret == SOCKET_ERROR) {
        log_printf("bind for IPv6 multicast failed: %ld\n", WSAGetLastError());
        cleanup();
        return 7;
    }

// below block for TTL setting should only be used on beacons, as it prevents
// use over the Internet XXX maybe this means beacons should have their own socket
	// set TTL on socket to 2 in case some ridiculous NAT decrements it
	//int ttl = 2;
	//ret = setsockopt(ipv4MulticastSocket, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&ttl, sizeof(ttl));
	//if (ret == SOCKET_ERROR) { // non-fatal to us
	//	printf("setsockopt() for multicast ttl failed: %ld\n", WSAGetLastError());
	//}

	// join multicast group(s)
	struct ip_mreq mreq;
	mreq.imr_multiaddr.s_addr = inet_addr(SARATOGA_IPV4_MULTICAST_ADDR);
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	ret = setsockopt(ipv4MulticastSocket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (char *)&mreq, sizeof(mreq));
	if (ret == SOCKET_ERROR) { // probably non-fatal to us ...
		log_printf("setsockopt() for multicast group join failed: %ld\n", WSAGetLastError());
	}
    struct ipv6_mreq mreq6;
    mreq6.ipv6mr_interface = 0;
    inet_pton(AF_INET6, SARATOGA_IPV6_MULTICAST_ADDR, &mreq6.ipv6mr_multiaddr);

	// configure multicast address used for sending
	ipv4MulticastAddress.sin_family = AF_INET;
	ipv4MulticastAddress.sin_addr.s_addr = inet_addr(SARATOGA_IPV4_MULTICAST_ADDR);
	ipv4MulticastAddress.sin_port = htons(SARATOGA_DEFAULT_PORT);

    ipv6MulticastAddress.sin6_family = AF_INET6;
    ipv6MulticastAddress.sin6_port = htons(SARATOGA_DEFAULT_PORT);
    inet_pton(AF_INET6, SARATOGA_IPV6_MULTICAST_ADDR, &ipv6MulticastAddress.sin6_addr);

	sendBeacon();

	return 0;
}

void cleanup() {
	int ret;

	// drop membership in multicast group
	if (ipv4MulticastSocket != SOCKET_ERROR) {
		struct ip_mreq mreq;
		mreq.imr_multiaddr.s_addr = inet_addr(SARATOGA_IPV4_MULTICAST_ADDR);
		mreq.imr_interface.s_addr = INADDR_ANY;
		ret = setsockopt(ipv4MulticastSocket, IPPROTO_IP, IP_DROP_MEMBERSHIP, (char *)&mreq, sizeof(mreq));
		if (ret == SOCKET_ERROR) {
			log_printf("setsockopt() for multicast group leave failed: %ld\n", WSAGetLastError());
		}
	}

	if (ipv4UnicastSocket != SOCKET_ERROR) closesocket(ipv4UnicastSocket);
	if (ipv4MulticastSocket != SOCKET_ERROR) closesocket(ipv4MulticastSocket);

	WSACleanup();

    log_printf("closing log\n");
	if (logFP != NULL) fclose(logFP);
    logFP = NULL;
}

void serviceSocket(SOCKET s) {
	const int packetBufLen = 4096;
	char packetBuf[packetBufLen];
	struct sockaddr_storage fromAddr;
	socklen_t fromAddrLen = sizeof(fromAddr);
	int packetLen = recvfrom(s, packetBuf, packetBufLen, 0, (struct sockaddr *)&fromAddr, &fromAddrLen);
	if (packetLen > 0) {
		handlePacket(packetBuf, packetLen, &fromAddr);
	} else printf("recvfrom() failed: %ld\n", WSAGetLastError());
}

bool processIncomingPackets() {
printf("processIncomingPackets\n");
	struct timeval tv;
	tv.tv_sec = tv.tv_usec = 0;
	fd_set rfds;
	FD_ZERO(&rfds);
	FD_SET(ipv4UnicastSocket, &rfds);
	FD_SET(ipv4MulticastSocket, &rfds);
    FD_SET(ipv6UnicastSocket, &rfds);
    FD_SET(ipv6MulticastSocket, &rfds);
    int socketsActive = 0;
	if ((socketsActive = select(0, &rfds, NULL, NULL, &tv)) > 0) {
		if (FD_ISSET(ipv4UnicastSocket, &rfds)) {
			serviceSocket(ipv4UnicastSocket);
		}
		if (FD_ISSET(ipv4MulticastSocket, &rfds)) {
			serviceSocket(ipv4MulticastSocket);
		}
        if (FD_ISSET(ipv6UnicastSocket, &rfds)) {
            serviceSocket(ipv6UnicastSocket);
        }
        if (FD_ISSET(ipv6MulticastSocket, &rfds)) {
            serviceSocket(ipv6MulticastSocket);
        }
	}
printf("returning\n");
    return socketsActive != 0;
}

};

